import atexit
import json
import os
import subprocess
import time
from distutils import spawn

import pytest
import responses
from dagster_dbt.resources import DbtRpcClient
from six.moves.urllib import request
from six.moves.urllib.error import URLError

from dagster.utils import file_relative_path, pushd
from dagster.utils.test.postgres_instance import TestPostgresInstance

# ======= CONFIG ========
DBT_EXECUTABLE = "dbt"
TEST_PROJECT_DIR = file_relative_path(__file__, "dagster_dbt_test_project")
DBT_CONFIG_DIR = os.path.join(TEST_PROJECT_DIR, "dbt_config")
TEST_HOSTNAME = "127.0.0.1"
TEST_PORT = 8580
RPC_ESTABLISH_RETRIES = 4
RPC_ESTABLISH_RETRY_INTERVAL_S = 1.5

RPC_ENDPOINT = "http://{hostname}:{port}/jsonrpc".format(hostname=TEST_HOSTNAME, port=TEST_PORT)

# ======= SOLIDS I ========
@pytest.fixture(scope="session")
def conn_string():
    postgres_host = os.environ.get("POSTGRES_TEST_DB_DBT_HOST")
    if not postgres_host:
        pytest.fail("Env variable POSTGRES_TEST_DB_DBT_HOST is unset")

    with TestPostgresInstance.docker_service_up_or_skip(
        file_relative_path(__file__, "docker-compose.yml"),
        "test-postgres-db-dbt",
        {"hostname": postgres_host},
    ) as conn_str:
        yield conn_str


@pytest.fixture(scope="session")
def prepare_dbt_cli(conn_string):  # pylint: disable=unused-argument, redefined-outer-name
    if not spawn.find_executable(DBT_EXECUTABLE):
        raise Exception("executable not found in path for `dbt`")

    with pushd(TEST_PROJECT_DIR):
        yield


@pytest.fixture(scope="class")
def dbt_seed(prepare_dbt_cli):  # pylint: disable=unused-argument, redefined-outer-name
    subprocess.run([DBT_EXECUTABLE, "seed", "--profiles-dir", DBT_CONFIG_DIR], check=True)


def get_rpc_server_status():
    status_request_body = b'{"jsonrpc": "2.0", "method": "status", "id": 1}'
    req = request.Request(
        RPC_ENDPOINT, data=status_request_body, headers={"Content-type": "application/json"},
    )
    resp = request.urlopen(req)
    return json.load(resp)


all_subprocs = set()


def kill_all_subprocs():
    for proc in all_subprocs:
        proc.kill()


atexit.register(kill_all_subprocs)


@pytest.fixture(scope="class")
def dbt_rpc_server(dbt_seed):  # pylint: disable=unused-argument, redefined-outer-name
    proc = subprocess.Popen(
        [
            DBT_EXECUTABLE,
            "rpc",
            "--host",
            TEST_HOSTNAME,
            "--port",
            str(TEST_PORT),
            "--profiles-dir",
            DBT_CONFIG_DIR,
        ],
    )

    # schedule to be killed in case of abort
    all_subprocs.add(proc)

    tries_remaining = RPC_ESTABLISH_RETRIES
    while True:
        poll_result = proc.poll()  # check on the child
        if poll_result != None:
            # terminated early!?
            raise Exception("DBT subprocess terminated before test could start.")

        try:
            status_json = get_rpc_server_status()
            if status_json["result"]["state"] == "ready":
                break
        except URLError:
            pass

        if tries_remaining <= 0:
            raise Exception("Exceeded max tries waiting for DBT RPC server to be ready.")
        tries_remaining -= 1
        time.sleep(RPC_ESTABLISH_RETRY_INTERVAL_S)

    yield

    proc.terminate()  # clean up after ourself
    proc.wait(timeout=0.2)
    if proc.poll() == None:  # still running
        proc.kill()
    all_subprocs.remove(proc)


# ======= SOLIDS II ========
@pytest.fixture(scope="session")
def rpc_endpoint():
    return RPC_ENDPOINT


@pytest.fixture
def rsps():
    with responses.RequestsMock() as req_mock:
        yield req_mock


@pytest.fixture
def non_terminal_poll_result(rpc_logs):  # pylint: disable=redefined-outer-name
    result = {
        "result": {
            "state": "running",
            "start": "2020-03-10T17:49:39.095678Z",
            "end": None,
            "elapsed": 1.471953,
            "logs": rpc_logs,
            "tags": {},
        },
        "id": "846157fe-62f7-11ea-9cd7-acde48001122",
        "jsonrpc": "2.0",
    }
    return result


@pytest.fixture
def terminal_poll_result(rpc_logs):  # pylint: disable=redefined-outer-name
    result = {
        "result": {
            "state": "success",
            "start": "2020-03-10T17:52:19.254197Z",
            "end": "2020-03-10T17:53:06.195224Z",
            "elapsed": 46.941027,
            "logs": rpc_logs,
            "tags": {},
            "results": [
                {
                    "node": {
                        "raw_sql": "\n\n{{\n  config(\n    unique_key='ds',\n    strategy='check',\n    check_cols='all'\n  )\n}}\n\nselect src.*\nfrom {{ source('dagster', 'daily_fulfillment_forecast') }} src\n\n",
                        "database": "snapshots_david_wallace",
                        "schema": "dagster",
                        "fqn": [
                            "dataland_dbt",
                            "dagster",
                            "daily_fulfillment_forecast_snapshot",
                            "daily_fulfillment_forecast_snapshot",
                        ],
                        "unique_id": "snapshot.dataland_dbt.daily_fulfillment_forecast_snapshot",
                        "package_name": "dataland_dbt",
                        "root_path": "/Users/dwall/repos/dataland-dbt",
                        "path": "dagster/daily_fulfillment_forecast_snapshot.sql",
                        "original_file_path": "snapshots/dagster/daily_fulfillment_forecast_snapshot.sql",
                        "name": "daily_fulfillment_forecast_snapshot",
                        "resource_type": "snapshot",
                        "alias": "daily_fulfillment_forecast_snapshot",
                        "config": {
                            "enabled": True,
                            "materialized": "snapshot",
                            "persist_docs": {},
                            "post-hook": [],
                            "pre-hook": [],
                            "vars": {},
                            "quoting": {},
                            "column_types": {},
                            "tags": [],
                            "unique_key": "ds",
                            "target_schema": "dagster",
                            "target_database": "snapshots_david_wallace",
                            "strategy": "check",
                            "check_cols": "all",
                            "transient": False,
                        },
                        "tags": [],
                        "refs": [],
                        "sources": [["dagster", "daily_fulfillment_forecast"]],
                        "depends_on": {
                            "nodes": ["source.dataland_dbt.dagster.daily_fulfillment_forecast"],
                            "macros": [],
                        },
                        "docrefs": [],
                        "description": "",
                        "columns": {},
                        "patch_path": None,
                        "build_path": "target/run/dataland_dbt/dagster/daily_fulfillment_forecast_snapshot.sql",
                        "compiled": True,
                        "compiled_sql": "\n\n\n\nselect src.*\nfrom ingest_dev.dagster.daily_fulfillment_forecast src\n",
                        "extra_ctes_injected": True,
                        "extra_ctes": [],
                        "injected_sql": "\n\n\n\nselect src.*\nfrom ingest_dev.dagster.daily_fulfillment_forecast src\n",
                        "wrapped_sql": "None",
                    },
                    "error": None,
                    "status": "SUCCESS 0",
                    "execution_time": 14.527844190597534,
                    "thread_id": "Thread-1",
                    "timing": [
                        {
                            "name": "compile",
                            "started_at": "2020-03-10T17:52:50.519541Z",
                            "completed_at": "2020-03-10T17:52:50.533709Z",
                        },
                        {
                            "name": "execute",
                            "started_at": "2020-03-10T17:52:50.533986Z",
                            "completed_at": "2020-03-10T17:53:05.046646Z",
                        },
                    ],
                    "fail": None,
                    "warn": None,
                    "skip": False,
                }
            ],
            "generated_at": "2020-03-10T17:53:06.001341Z",
            "elapsed_time": 44.305715799331665,
        },
        "id": "016d2822-62f8-11ea-906b-acde48001122",
        "jsonrpc": "2.0",
    }
    return result


# ======= CLIENT ========
@pytest.fixture(scope="session")
def client():
    return DbtRpcClient(host="0.0.0.0", port=8580)


# ======= UTILS ========
@pytest.fixture
def rpc_logs():
    return [
        {
            "timestamp": "2020-03-10T18:19:06.726848Z",
            "message": "finished collecting timing info",
            "channel": "dbt",
            "level": 10,
            "levelname": "DEBUG",
            "thread_name": "Thread-1",
            "process": 18546,
            "extra": {
                "timing_info": {
                    "name": "execute",
                    "started_at": "2020-03-10T18:18:54.823894Z",
                    "completed_at": "2020-03-10T18:19:06.726805Z",
                },
                "json_only": True,
                "unique_id": "snapshot.dataland_dbt.daily_fulfillment_forecast_snapshot",
                "run_state": "running",
                "context": "server",
            },
            "exc_info": None,
        },
        {
            "timestamp": "2020-03-10T18:19:06.727723Z",
            "message": "11:19:06 | 1 of 1 OK snapshotted snapshots_david_wallace.dagster.daily_fulfillment_forecast_snapshot [\u001b[32mSUCCESS 0\u001b[0m in 11.92s]",
            "channel": "dbt",
            "level": 11,
            "levelname": "INFO",
            "thread_name": "Thread-1",
            "process": 18546,
            "extra": {
                "unique_id": "snapshot.dataland_dbt.daily_fulfillment_forecast_snapshot",
                "run_state": "running",
                "context": "server",
            },
            "exc_info": None,
        },
    ]



from dagster.core.utils import check_dagster_package_version

from .resources import dbt_rpc_resource, local_dbt_rpc_resource
from .solids import (
    create_dbt_rpc_run_sql_solid,
    dbt_rpc_compile_sql,
    dbt_rpc_run,
    dbt_rpc_run_and_wait,
    dbt_rpc_run_operation,
    dbt_rpc_run_operation_and_wait,
    dbt_rpc_snapshot,
    dbt_rpc_snapshot_and_wait,
    dbt_rpc_snapshot_freshness,
    dbt_rpc_snapshot_freshness_and_wait,
    dbt_rpc_test,
    dbt_rpc_test_and_wait,
)
from .types import DbtRpcPollResult
from .version import __version__

check_dagster_package_version("dagster-dbt", __version__)

__all__ = [
    "DbtRpcPollResult",
    "dbt_rpc_resource",
    "dbt_rpc_run",
    "dbt_rpc_run_and_wait",
    "dbt_rpc_test",
    "create_dbt_rpc_run_sql_solid",
    "dbt_rpc_compile_sql",
    "dbt_rpc_run",
    "dbt_rpc_run_and_wait",
    "dbt_rpc_run_operation",
    "dbt_rpc_run_operation_and_wait",
    "dbt_rpc_snapshot",
    "dbt_rpc_snapshot_and_wait",
    "dbt_rpc_snapshot_freshness",
    "dbt_rpc_snapshot_freshness_and_wait",
    "dbt_rpc_test",
    "dbt_rpc_test_and_wait",
    "local_dbt_rpc_resource",
]



from datetime import datetime, timedelta
from typing import Dict, List, Optional

import attr
import pytz
from dateutil import parser

from dagster import usable_as_dagster_type


def datetime_converter(value):
    if isinstance(value, datetime):
        return value
    elif value is not None:
        return parser.isoparse(value).replace(
            tzinfo=pytz.UTC
        )  # TODO does this need to be dealt with differently in py27? see https://dagster.phacility.com/D2295
    else:
        return None


@attr.s
class NodeTiming(object):
    _name: str = attr.ib()
    _started_at: datetime = attr.ib(converter=datetime_converter)
    _completed_at: datetime = attr.ib(converter=datetime_converter)

    @property
    def name(self):
        return self._name

    @property
    def started_at(self):
        return self._started_at

    @property
    def completed_at(self):
        return self._completed_at

    @property
    def duration(self) -> timedelta:
        return self._completed_at - self._started_at


@attr.s
class NodeResult(object):
    _node: Dict = attr.ib()
    _error: Optional[bool] = attr.ib()
    _status: str = attr.ib()
    _execution_time: float = attr.ib()
    _thread_id: str = attr.ib()
    _timing: List[NodeTiming] = attr.ib()
    _fail: Optional[bool] = attr.ib()
    _warn: Optional[bool] = attr.ib()
    _skip: Optional[bool] = attr.ib()
    _table: Optional[Dict] = attr.ib(default=None)

    @classmethod
    def from_result(cls, result):
        return cls(
            node=result.get("node"),
            error=result.get("error"),
            status=result.get("status"),
            execution_time=result.get("execution_time"),
            thread_id=result.get("thread_id"),
            timing=[NodeTiming(**step) for step in result.get("timing")],
            fail=result.get("fail"),
            warn=result.get("warn"),
            skip=result.get("skip"),
            table=result.get("table"),
        )

    @property
    def node(self):
        return self._node

    @property
    def table(self):
        return self._table

    @property
    def status(self):
        return self._status

    @property
    def error(self):
        return self._error

    @property
    def fail(self):
        return self._fail

    @property
    def skip(self):
        return self._skip

    @property
    def timing(self):
        return self._timing

    @property
    def execution_time(self):
        return self._execution_time


@attr.s
@usable_as_dagster_type
class DbtRpcPollResult(object):
    _state: str = attr.ib()
    _start: datetime = attr.ib(converter=datetime_converter)
    _end: datetime = attr.ib(converter=datetime_converter)
    _elapsed: float = attr.ib()
    _logs: List[Dict] = attr.ib()
    _tags: Dict = attr.ib()
    _results: List[NodeResult] = attr.ib(factory=list)
    _generated_at: Optional[datetime] = attr.ib(converter=datetime_converter, default=None)
    _elapsed_time: Optional[float] = attr.ib(default=None)
    _success: Optional[bool] = attr.ib(default=None)

    @classmethod
    def from_results(cls, results):

        node_results = []
        if results.get("results") is not None:
            for result in results.get("results"):
                node_results.append(NodeResult.from_result(result))

        return cls(
            state=results.get("state"),
            start=results.get("start"),
            end=results.get("end"),
            elapsed=results.get("elapsed"),
            logs=results.get("logs"),
            tags=results.get("tags"),
            results=node_results,
            generated_at=results.get("generated_at"),
            elapsed_time=results.get("elapsed_time"),
            success=results.get("success"),
        )

    @property
    def results(self):
        return self._results

    def __len__(self):
        return len(self._results)

    def __getitem__(self, position):
        return self._results[position]



import pytest
import responses
from dagster_dbt.resources import DbtRpcClient, dbt_rpc_resource, local_dbt_rpc_resource

from dagster import ModeDefinition, execute_solid, solid


def test_url(client):
    assert client.url == "http://0.0.0.0:8580/jsonrpc"


@pytest.mark.parametrize("method", ["status", "poll", "kill", "cli_args"])
def test_default_request(client, method):
    expected = {"jsonrpc": client.jsonrpc_version, "method": method, "params": {}}
    resp = client._default_request(method=method)  # pylint: disable=protected-access
    assert resp["jsonrpc"] == expected["jsonrpc"]
    assert resp["method"] == expected["method"]
    assert resp["params"] == expected["params"]


def test_selection(client):
    expected = {
        "models": "@model_1 +model_2+ model_3+",
        "select": "snapshot_1 snapshot_2 snapshot_3",
        "exclude": "model_4+",
    }
    data = client._selection(  # pylint: disable=protected-access
        models=["@model_1", "+model_2+", "model_3+", "model_3+"],
        select=["snapshot_1", "snapshot_2", "snapshot_3"],
        exclude=["model_4+"],
    )

    assert set(data["models"].split(" ")) == set(expected["models"].split(" "))
    assert set(data["select"].split(" ")) == set(expected["select"].split(" "))
    assert set(data["exclude"].split(" ")) == set(expected["exclude"].split(" "))


def test_status(client):
    with responses.RequestsMock() as rsps:
        expected = {
            "result": {
                "status": "ready",
                "error": "null",
                "logs": [],
                "timestamp": "2019-10-07T16:30:09.875534Z",
                "pid": 76715,
            },
            "id": "2db9a2fe-9a39-41ef-828c-25e04dd6b07d",
            "jsonrpc": client.jsonrpc_version,
        }
        rsps.add(responses.POST, client.url, json=expected, status=202)

        resp = client.status()
        assert resp.json() == expected


def test_poll(client):
    with responses.RequestsMock() as rsps:
        expected = {
            "result": {
                "results": [],
                "generated_at": "2019-10-11T18:25:22.477203Z",
                "elapsed_time": 0.8381369113922119,
                "logs": [],
                "tags": {"command": "run --models my_model", "branch": "abc123"},
                "status": "success",
            },
            "id": "2db9a2fe-9a39-41ef-828c-25e04dd6b07d",
            "jsonrpc": client.jsonrpc_version,
        }
        rsps.add(responses.POST, client.url, json=expected, status=202)

        resp = client.poll(request_token="f86926fa-6535-4891-8d24-2cfc65d2a347")
        assert resp.json() == expected


def test_dbt_rpc_resource():
    it = {}

    @solid(required_resource_keys={"dbt_rpc"})
    def a_solid(context):
        assert isinstance(context.resources.dbt_rpc, DbtRpcClient)
        assert context.resources.dbt_rpc.host == "<default host>"
        assert context.resources.dbt_rpc.port == 8580
        it["ran"] = True

    execute_solid(
        a_solid,
        ModeDefinition(resource_defs={"dbt_rpc": dbt_rpc_resource}),
        None,
        None,
        {"resources": {"dbt_rpc": {"config": {"host": "<default host>"}}}},
    )
    assert it["ran"]


def test_local_dbt_rpc_resource():
    it = {}

    @solid(required_resource_keys={"dbt_rpc"})
    def a_solid(context):
        assert isinstance(context.resources.dbt_rpc, DbtRpcClient)
        assert context.resources.dbt_rpc.host == "0.0.0.0"
        assert context.resources.dbt_rpc.port == 8580
        it["ran"] = True

    execute_solid(a_solid, ModeDefinition(resource_defs={"dbt_rpc": local_dbt_rpc_resource}))
    assert it["ran"]



import json
import time
from typing import Callable, Iterator, Optional

import pandas as pd
from dagster_pandas import DataFrame

from dagster import (
    Array,
    AssetMaterialization,
    Bool,
    EventMetadataEntry,
    Failure,
    Field,
    InputDefinition,
    Int,
    Noneable,
    Nothing,
    Output,
    OutputDefinition,
    Permissive,
    RetryRequested,
    String,
    check,
    solid,
)

from .types import DbtRpcPollResult
from .utils import log_rpc, raise_for_rpc_error


def generate_materializations(rpc_poll_result: DbtRpcPollResult) -> Iterator[AssetMaterialization]:
    for node_result in rpc_poll_result.results:
        if node_result.node["resource_type"] in ["model", "snapshot"]:
            success = not node_result.fail and not node_result.skip and not node_result.error
            if success:
                entries = [
                    EventMetadataEntry.json(data=node_result.node, label="Node"),
                    EventMetadataEntry.text(text=node_result.status, label="Status"),
                    EventMetadataEntry.text(
                        text=str(node_result.execution_time), label="Execution Time"
                    ),
                    EventMetadataEntry.text(
                        text=node_result.node["config"]["materialized"],
                        label="Materialization Strategy",
                    ),
                    EventMetadataEntry.text(text=node_result.node["database"], label="Database"),
                    EventMetadataEntry.text(text=node_result.node["schema"], label="Schema"),
                    EventMetadataEntry.text(text=node_result.node["alias"], label="Alias"),
                    EventMetadataEntry.text(
                        text=node_result.node["description"], label="Description"
                    ),
                ]
                for node_timing in node_result.timing:
                    if node_timing.name == "execute":
                        execution_entries = [
                            EventMetadataEntry.text(
                                text=node_timing.started_at.isoformat(timespec="seconds"),
                                label="Execution Started At",
                            ),
                            EventMetadataEntry.text(
                                text=node_timing.completed_at.isoformat(timespec="seconds"),
                                label="Execution Completed At",
                            ),
                            EventMetadataEntry.text(
                                text=str(node_timing.duration), label="Execution Duration"
                            ),
                        ]
                        entries.extend(execution_entries)
                    if node_timing.name == "compile":
                        execution_entries = [
                            EventMetadataEntry.text(
                                text=node_timing.started_at.isoformat(timespec="seconds"),
                                label="Compilation Started At",
                            ),
                            EventMetadataEntry.text(
                                text=node_timing.completed_at.isoformat(timespec="seconds"),
                                label="Compilation Completed At",
                            ),
                            EventMetadataEntry.text(
                                text=str(node_timing.duration), label="Compilation Duration"
                            ),
                        ]
                        entries.extend(execution_entries)

                yield AssetMaterialization(
                    description="A materialized node within the dbt graph.",
                    metadata_entries=entries,
                    asset_key=node_result.node["unique_id"],
                )


def dbt_rpc_poll(
    context, request_token: str, should_yield_materializations: bool = True
) -> DbtRpcPollResult:

    resp = context.resources.dbt_rpc.poll(
        request_token=request_token, logs=context.solid_config["logs"]
    )
    raise_for_rpc_error(context, resp)

    interval = context.solid_config["interval"]
    logs_start = 0
    while resp.json().get("result").get("state") == "running":
        context.log.debug(
            f"Request {request_token} currently in state '{resp.json().get('result').get('state')}' (elapsed time {resp.json().get('result').get('elapsed', 0)} seconds). Sleeping for {interval}s.."
        )

        if context.solid_config["logs"]:
            logs = resp.json().get("result").get("logs")
            if len(logs) > 0:
                log_rpc(context, logs)
            logs_start += len(logs)

        time.sleep(interval)
        resp = context.resources.dbt_rpc.poll(
            request_token=request_token, logs=context.solid_config["logs"], logs_start=logs_start
        )
        raise_for_rpc_error(context, resp)

    if resp.json().get("result").get("state") != "success":
        if context.solid_config["logs"]:
            logs = resp.json().get("result").get("logs")
            if len(logs) > 0:
                log_rpc(context, logs)

        raise Failure(
            description=f"Request {request_token} finished with state '{resp.json().get('result').get('state')}' in {resp.json().get('result').get('elapsed')} seconds",
        )

    else:
        context.log.info(
            f"Request {request_token} finished with state '{resp.json().get('result').get('state')}' in {resp.json().get('result').get('elapsed')} seconds"
        )

    if context.solid_config["logs"]:
        logs = resp.json().get("result").get("logs")
        if len(logs) > 0:
            log_rpc(context, logs)

    context.log.debug(json.dumps(resp.json().get("result"), indent=2))
    rpc_poll_result = DbtRpcPollResult.from_results(resp.json().get("result"))
    if should_yield_materializations:
        for materialization in generate_materializations(rpc_poll_result=rpc_poll_result):
            yield materialization
    yield Output(value=rpc_poll_result, output_name="result")


@solid(
    description="A solid to invoke dbt run over RPC.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[
        OutputDefinition(
            name="request_token",
            dagster_type=String,
            description="The request token of the invoked dbt run.",
        )
    ],
    config_schema={
        "models": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to run.",
        ),
        "exclude": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to exclude.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_run(context) -> String:
    resp = context.resources.dbt_rpc.run(
        models=context.solid_config["models"], exclude=context.solid_config["exclude"]
    )
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    return resp.json().get("result").get("request_token")


@solid(
    description="A solid to invoke dbt run over RPC and poll the resulting RPC process until it's complete.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[OutputDefinition(name="result", dagster_type=DbtRpcPollResult)],
    config_schema={
        "models": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to run.",
        ),
        "exclude": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to exclude.",
        ),
        "full_refresh": Field(
            config=Bool,
            description="Whether or not to perform a --full-refresh.",
            is_required=False,
            default_value=False,
        ),
        "fail_fast": Field(
            config=Bool,
            description="Whether or not to --fail-fast.",
            is_required=False,
            default_value=False,
        ),
        "warn_error": Field(
            config=Bool,
            description="Whether or not to --warn-error.",
            is_required=False,
            default_value=False,
        ),
        "interval": Field(
            config=Int,
            is_required=False,
            default_value=10,
            description="The interval (in seconds) at which to poll the dbt rpc process.",
        ),
        "logs": Field(
            config=Bool,
            is_required=False,
            default_value=True,
            description="Whether or not to return logs from the process.",
        ),
        "task_tags": Permissive(),
        "max_retries": Field(config=Int, is_required=False, default_value=5),
        "retry_interval": Field(config=Int, is_required=False, default_value=120),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_run_and_wait(context) -> DbtRpcPollResult:

    if context.solid_config["task_tags"]:
        results = context.resources.dbt_rpc.ps().json()
        for task in results["result"]["rows"]:
            if task["tags"] == context.solid_config["task_tags"]:
                context.log.warning(
                    f"RPC task with tags {json.dumps(task['tags'])} currently running."
                )
                raise RetryRequested(
                    max_retries=context.solid_config["max_retries"],
                    seconds_to_wait=context.solid_config["retry_interval"],
                )

    command = ""

    if context.solid_config["warn_error"]:
        command += " --warn-error"

    command += " run"

    if context.solid_config["models"]:
        models = " ".join(set(context.solid_config["models"]))
        command += f" --models {models}"

    if context.solid_config["exclude"]:
        exclude = " ".join(set(context.solid_config["exclude"]))
        command += f" --exclude {exclude}"

    if context.solid_config["full_refresh"]:
        command += " --full-refresh"

    if context.solid_config["fail_fast"]:
        command += " --fail-fast"

    context.log.debug(f"Running dbt command: dbt {command}")
    resp = context.resources.dbt_rpc.cli(cli=command, **context.solid_config["task_tags"])
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    request_token = resp.json().get("result").get("request_token")
    return dbt_rpc_poll(context, request_token)


@solid(
    description="A solid to invoke dbt test over RPC.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[
        OutputDefinition(
            name="request_token",
            dagster_type=String,
            description="The request token of the invoked dbt test.",
        )
    ],
    config_schema={
        "models": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to test.",
        ),
        "exclude": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to exclude.",
        ),
        "data": Field(
            config=Bool,
            default_value=True,
            is_required=False,
            description="Whether or not to run custom data tests.",
        ),
        "schema": Field(
            config=Bool,
            default_value=True,
            is_required=False,
            description="Whether or not to run schema tests.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_test(context) -> String:
    resp = context.resources.dbt_rpc.test(
        models=context.solid_config["models"],
        exclude=context.solid_config["exclude"],
        data=context.solid_config["data"],
        schema=context.solid_config["schema"],
    )
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    return resp.json().get("result").get("request_token")


@solid(
    description="A solid to invoke dbt test over RPC and poll the resulting RPC process until it's complete.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[OutputDefinition(name="result", dagster_type=DbtRpcPollResult)],
    config_schema={
        "models": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to test.",
        ),
        "exclude": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt models to exclude.",
        ),
        "data": Field(
            config=Bool,
            default_value=True,
            is_required=False,
            description="Whether or not to run custom data tests.",
        ),
        "schema": Field(
            config=Bool,
            default_value=True,
            is_required=False,
            description="Whether or not to run schema tests.",
        ),
        "interval": Field(
            config=Int,
            is_required=False,
            default_value=10,
            description="The interval (in seconds) at which to poll the dbt rpc process.",
        ),
        "logs": Field(
            config=Bool,
            is_required=False,
            default_value=True,
            description="Whether or not to return logs from the process.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_test_and_wait(context) -> DbtRpcPollResult:
    resp = context.resources.dbt_rpc.test(
        models=context.solid_config["models"],
        exclude=context.solid_config["exclude"],
        data=context.solid_config["data"],
        schema=context.solid_config["schema"],
    )
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    request_token = resp.json().get("result").get("request_token")
    return dbt_rpc_poll(context, request_token)


@solid(
    description="A solid to invoke a dbt run operation over RPC.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[
        OutputDefinition(
            name="request_token",
            dagster_type=String,
            description="The request token of the invoked dbt run operation.",
        )
    ],
    config_schema={
        "macro": Field(
            config=String,
            is_required=True,
            description="The dbt macro to invoke as a run operation",
        ),
        "args": Field(
            config=Noneable(Permissive()),
            is_required=False,
            default_value=None,
            description="Arguments to supply to the invoked macro.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_run_operation(context) -> String:
    resp = context.resources.dbt_rpc.run_operation(
        macro=context.solid_config["macro"], args=context.solid_config["args"]
    )
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    return resp.json().get("result").get("request_token")


@solid(
    description="A solid to invoke a dbt run operation over RPC and poll the resulting RPC process until it's complete.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[OutputDefinition(name="result", dagster_type=DbtRpcPollResult)],
    config_schema={
        "macro": Field(
            config=String,
            is_required=True,
            description="The dbt macro to invoke as a run operation",
        ),
        "args": Field(
            config=Noneable(Permissive()),
            is_required=False,
            default_value=None,
            description="Arguments to supply to the invoked macro.",
        ),
        "interval": Field(
            config=Int,
            is_required=False,
            default_value=10,
            description="The interval (in seconds) at which to poll the dbt rpc process.",
        ),
        "logs": Field(
            config=Bool,
            is_required=False,
            default_value=True,
            description="Whether or not to return logs from the process.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_run_operation_and_wait(context) -> DbtRpcPollResult:
    resp = context.resources.dbt_rpc.run_operation(
        macro=context.solid_config["macro"], args=context.solid_config["args"]
    )
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    request_token = resp.json().get("result").get("request_token")
    return dbt_rpc_poll(context, request_token)


@solid(
    description="A solid to invoke a dbt snapshot over RPC.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[
        OutputDefinition(
            name="request_token",
            dagster_type=String,
            description="The request token of the invoked dbt snapshot.",
        )
    ],
    config_schema={
        "select": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt snapshot files to snapshot.",
        ),
        "exclude": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt snapshot files to exclude from the snapshot.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_snapshot(context) -> String:
    resp = context.resources.dbt_rpc.snapshot(
        select=context.solid_config["select"], exclude=context.solid_config["exclude"]
    )
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    return resp.json().get("result").get("request_token")


@solid(
    description="A solid to invoke a dbt snapshot over RPC and poll the resulting RPC process until it's complete.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[OutputDefinition(name="result", dagster_type=DbtRpcPollResult)],
    config_schema={
        "select": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt snapshot files to snapshot.",
        ),
        "exclude": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt snapshot files to exclude from the snapshot.",
        ),
        "interval": Field(
            config=Int,
            is_required=False,
            default_value=10,
            description="The interval (in seconds) at which to poll the dbt rpc process.",
        ),
        "logs": Field(
            config=Bool,
            is_required=False,
            default_value=True,
            description="Whether or not to return logs from the process.",
        ),
        "task_tags": Permissive(),
        "max_retries": Field(config=Int, is_required=False, default_value=5),
        "retry_interval": Field(config=Int, is_required=False, default_value=120),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_snapshot_and_wait(context) -> DbtRpcPollResult:

    if context.solid_config["task_tags"]:
        results = context.resources.dbt_rpc.ps().json()
        for task in results["result"]["rows"]:
            if task["tags"] == context.solid_config["task_tags"]:
                context.log.warning(
                    f"RPC task with tags {json.dumps(task['tags'])} currently running."
                )
                raise RetryRequested(
                    max_retries=context.solid_config["max_retries"],
                    seconds_to_wait=context.solid_config["retry_interval"],
                )

    resp = context.resources.dbt_rpc.snapshot(
        select=context.solid_config["select"], exclude=context.solid_config["exclude"]
    )
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    request_token = resp.json().get("result").get("request_token")
    return dbt_rpc_poll(context, request_token)


@solid(
    description="A solid to invoke dbt source snapshot-freshness over RPC.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[
        OutputDefinition(
            name="request_token",
            dagster_type=String,
            description="The request token of the invoked dbt snapshot.",
        )
    ],
    config_schema={
        "select": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt sources to snapshot-freshness for.",
        ),
        "warn_error": Field(
            config=Bool,
            description="Whether or not to --warn-error.",
            is_required=False,
            default_value=False,
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_snapshot_freshness(context) -> String:
    command = ""

    if context.solid_config["warn_error"]:
        command += " --warn-error"

    command += " source snapshot-freshness"

    if context.solid_config["select"]:
        select = " ".join(set(context.solid_config["select"]))
        command += f" --select {select}"

    context.log.debug(f"Running dbt command: dbt {command}")
    resp = context.resources.dbt_rpc.cli(cli=command)
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    return resp.json().get("result").get("request_token")


@solid(
    description="A solid to invoke dbt source snapshot-freshness over RPC and poll the resulting RPC process until it's complete.",
    input_defs=[InputDefinition(name="start_after", dagster_type=Nothing)],
    output_defs=[OutputDefinition(name="result", dagster_type=DbtRpcPollResult)],
    config_schema={
        "select": Field(
            config=Noneable(Array(String)),
            default_value=None,
            is_required=False,
            description="The dbt sources to snapshot-freshness for.",
        ),
        "warn_error": Field(
            config=Bool,
            description="Whether or not to --warn-error.",
            is_required=False,
            default_value=False,
        ),
        "interval": Field(
            config=Int,
            is_required=False,
            default_value=10,
            description="The interval (in seconds) at which to poll the dbt rpc process.",
        ),
        "logs": Field(
            config=Bool,
            is_required=False,
            default_value=True,
            description="Whether or not to return logs from the process.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_snapshot_freshness_and_wait(context) -> DbtRpcPollResult:
    command = ""

    if context.solid_config["warn_error"]:
        command += " --warn-error"

    command += " source snapshot-freshness"

    if context.solid_config["select"]:
        select = " ".join(set(context.solid_config["select"]))
        command += f" --select {select}"

    context.log.debug(f"Running dbt command: dbt {command}")
    resp = context.resources.dbt_rpc.cli(cli=command)
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    request_token = resp.json().get("result").get("request_token")
    return dbt_rpc_poll(context, request_token)


@solid(
    description="A solid to compile a SQL query in context of a dbt project over RPC.",
    input_defs=[
        InputDefinition(name="start_after", dagster_type=Nothing),
        InputDefinition(
            name="sql", description="The SQL query to be compiled.", dagster_type=String
        ),
    ],
    output_defs=[
        OutputDefinition(name="sql", description="The compiled SQL query.", dagster_type=String)
    ],
    config_schema={
        "name": Field(config=String),
        "interval": Field(
            config=Int,
            is_required=False,
            default_value=10,
            description="The interval (in seconds) at which to poll the dbt rpc process.",
        ),
        "logs": Field(
            config=Bool,
            is_required=False,
            default_value=True,
            description="Whether or not to return logs from the process.",
        ),
    },
    required_resource_keys={"dbt_rpc"},
    tags={"kind": "dbt"},
)
def dbt_rpc_compile_sql(context, sql: String) -> String:
    resp = context.resources.dbt_rpc.compile_sql(sql=sql, name=context.solid_config["name"])
    context.log.debug(resp.text)
    raise_for_rpc_error(context, resp)
    request_token = resp.json().get("result").get("request_token")
    result = dbt_rpc_poll(context, request_token)
    return result.results[0].node["compiled_sql"]  # pylint: disable=no-member # TODO


def create_dbt_rpc_run_sql_solid(
    name: str, output_def: Optional[OutputDefinition] = None, **kwargs
) -> Callable:
    """This function is a factory which constructs a solid that will copy the results of a SQL query run within the context of a dbt project to a DataFrame.

    Any kwargs passed to this function will be passed along to the underlying @solid decorator.
    However, note that overriding config, input_defs, and required_resource_keys is not supported. You might consider using
    @composite_solid to wrap this solid in the cases where you'd like to configure the solid
    with different config fields.

    Args:
        name (str): The name of this solid.
        output_def (OutputDefinition, optional): The OutputDefinition for the solid. This value should always be a representation
            of a pandas DataFrame. If not specificed, the solid will default to an OutputDefinition named "df" with a DataFrame dagster type.

    Returns:
        SolidDefinition: Returns the constructed solid definition.
    """
    check.str_param(obj=name, param_name="name")
    check.opt_inst_param(obj=output_def, param_name="output_def", ttype=OutputDefinition)
    check.param_invariant(
        "input_defs" not in kwargs, "input_defs", "Overriding input_defs is not supported."
    )
    check.param_invariant(
        "required_resource_keys" not in kwargs,
        "required_resource_keys",
        "Overriding required_resource_keys is not supported.",
    )

    @solid(
        name=name,
        description=kwargs.pop(
            "description",
            "A solid to run a SQL query in context of a dbt project over RPC and return the results in a pandas DataFrame.",
        ),
        input_defs=[
            InputDefinition(name="start_after", dagster_type=Nothing),
            InputDefinition(
                name="sql", description="The SQL query to be run.", dagster_type=String
            ),
        ],
        output_defs=[
            output_def
            or OutputDefinition(
                name="df", description="The results of the SQL query.", dagster_type=DataFrame
            )
        ],
        config_schema={
            "name": Field(config=String),
            "interval": Field(
                config=Int,
                is_required=False,
                default_value=10,
                description="The interval (in seconds) at which to poll the dbt rpc process.",
            ),
            "logs": Field(
                config=Bool,
                is_required=False,
                default_value=True,
                description="Whether or not to return logs from the process.",
            ),
        },
        required_resource_keys={"dbt_rpc"},
        tags={"kind": "dbt"},
        **kwargs,
    )
    def _dbt_rpc_run_sql(context, sql: String) -> DataFrame:
        resp = context.resources.dbt_rpc.run_sql(sql=sql, name=context.solid_config["name"])
        context.log.debug(resp.text)
        raise_for_rpc_error(context, resp)
        request_token = resp.json().get("result").get("request_token")
        result = dbt_rpc_poll(context, request_token)
        table = result.results[0].table  # pylint: disable=no-member  # TODO
        return pd.DataFrame.from_records(data=table["rows"], columns=table["column_names"])

    return _dbt_rpc_run_sql



import responses
from dagster_dbt import (
    DbtRpcPollResult,
    dbt_rpc_resource,
    dbt_rpc_run,
    dbt_rpc_run_and_wait,
    dbt_rpc_run_operation,
    dbt_rpc_run_operation_and_wait,
    dbt_rpc_snapshot,
    dbt_rpc_snapshot_and_wait,
    dbt_rpc_snapshot_freshness,
    dbt_rpc_snapshot_freshness_and_wait,
    dbt_rpc_test,
    dbt_rpc_test_and_wait,
)

from dagster import ModeDefinition, execute_solid


def test_dbt_rpc_snapshot(rsps):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    result = execute_solid(
        dbt_rpc_snapshot,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={"resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},},
    )

    assert result.success
    assert result.output_value("request_token") == "1234-xo-xo"


def test_dbt_rpc_run(rsps):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    result = execute_solid(
        dbt_rpc_run,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={"resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},},
    )

    assert result.success
    assert result.output_value("request_token") == "1234-xo-xo"


def test_dbt_rpc_test(rsps):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    result = execute_solid(
        dbt_rpc_test,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={"resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},},
    )

    assert result.success
    assert result.output_value("request_token") == "1234-xo-xo"


def test_dbt_rpc_run_operation(rsps):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    result = execute_solid(
        dbt_rpc_run_operation,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={
            "resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},
            "solids": {"dbt_rpc_run_operation": {"config": {"macro": "my_test_macro"}}},
        },
    )

    assert result.success
    assert result.output_value("request_token") == "1234-xo-xo"


def test_dbt_rpc_snapshot_freshness(rsps):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    result = execute_solid(
        dbt_rpc_snapshot_freshness,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={"resources": {"dbt_rpc": {"config": {"host": host, "port": port}}}},
    )

    assert result.success
    assert result.output_value("request_token") == "1234-xo-xo"


def test_dbt_rpc_run_and_wait(rsps, non_terminal_poll_result, terminal_poll_result):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=non_terminal_poll_result,
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=terminal_poll_result,
    )

    result = execute_solid(
        dbt_rpc_run_and_wait,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={
            "resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},
            "solids": {"dbt_rpc_run_and_wait": {"config": {"interval": 1}}},
        },
    )

    assert result.success
    assert isinstance(result.output_value("result"), DbtRpcPollResult)


def test_dbt_rpc_snapshot_and_wait(rsps, non_terminal_poll_result, terminal_poll_result):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=non_terminal_poll_result,
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=terminal_poll_result,
    )

    result = execute_solid(
        dbt_rpc_snapshot_and_wait,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={
            "resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},
            "solids": {"dbt_rpc_snapshot_and_wait": {"config": {"interval": 1}}},
        },
    )

    assert result.success
    assert isinstance(result.output_value("result"), DbtRpcPollResult)


def test_dbt_rpc_snapshot_freshness_and_wait(rsps, non_terminal_poll_result, terminal_poll_result):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=non_terminal_poll_result,
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=terminal_poll_result,
    )

    result = execute_solid(
        dbt_rpc_snapshot_freshness_and_wait,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={
            "resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},
            "solids": {"dbt_rpc_snapshot_freshness_and_wait": {"config": {"interval": 1}}},
        },
    )

    assert result.success
    assert isinstance(result.output_value("result"), DbtRpcPollResult)


def test_dbt_rpc_run_operation_and_wait(rsps, non_terminal_poll_result, terminal_poll_result):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=non_terminal_poll_result,
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=terminal_poll_result,
    )

    result = execute_solid(
        dbt_rpc_run_operation_and_wait,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={
            "resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},
            "solids": {
                "dbt_rpc_run_operation_and_wait": {"config": {"macro": "test_macro", "interval": 1}}
            },
        },
    )

    assert result.success
    assert isinstance(result.output_value("result"), DbtRpcPollResult)


def test_dbt_rpc_test_and_wait(rsps, non_terminal_poll_result, terminal_poll_result):
    host = "0.0.0.0"
    port = 8580

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json={"result": {"request_token": "1234-xo-xo"}},
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=non_terminal_poll_result,
    )

    rsps.add(
        method=responses.POST,
        url=f"http://{host}:{port}/jsonrpc",
        status=201,
        json=terminal_poll_result,
    )

    result = execute_solid(
        dbt_rpc_test_and_wait,
        mode_def=ModeDefinition(name="unittest", resource_defs={"dbt_rpc": dbt_rpc_resource}),
        input_values={"start_after": None},
        run_config={
            "resources": {"dbt_rpc": {"config": {"host": host, "port": port}}},
            "solids": {"dbt_rpc_test_and_wait": {"config": {"interval": 1}}},
        },
    )

    assert result.success
    assert isinstance(result.output_value("result"), DbtRpcPollResult)



from dagster_dbt.utils import fmt_rpc_logs


def test_fmt_rpc_logs(rpc_logs):
    expected = {
        10: "2020-03-10T18:19:06.726848Z - DEBUG - finished collecting timing info",
        20: "2020-03-10T18:19:06.727723Z - INFO - 11:19:06 | 1 of 1 OK snapshotted snapshots_david_wallace.dagster.daily_fulfillment_forecast_snapshot [\x1b[32mSUCCESS 0\x1b[0m in 11.92s]",
    }
    fmttd_rpc_logs = fmt_rpc_logs(rpc_logs)
    assert expected == fmttd_rpc_logs



import uuid
from typing import Any, Dict, Tuple

import pytest
from dagster_dbt import DbtRpcPollResult, dbt_rpc_run_and_wait, dbt_rpc_test, local_dbt_rpc_resource
from dagster_dbt.resources import DbtRpcClient
from dagster_dbt.solids import dbt_rpc_run
from mock import MagicMock

from dagster import (
    DagsterInstance,
    ModeDefinition,
    PipelineDefinition,
    SolidDefinition,
    SolidExecutionResult,
    configured,
    execute_pipeline,
    execute_solid,
    resource,
)


def output_for_solid_executed_with_rpc_resource(
    a_solid, rpc_resource=local_dbt_rpc_resource
) -> Tuple[SolidExecutionResult, DbtRpcPollResult]:
    mode_def = ModeDefinition(resource_defs={"dbt_rpc": rpc_resource})  # use config defaults
    solid_result = execute_solid(a_solid, mode_def)

    assert solid_result.success
    output = solid_result.output_value()
    assert isinstance(output, DbtRpcPollResult)
    return solid_result, output


class TestDBTRunAndWaitSolid:

    ALL_MODELS_KEY_SET = {
        "model.dagster_dbt_test_project.sort_by_calories",
        "model.dagster_dbt_test_project.least_caloric",
        "model.dagster_dbt_test_project.sort_cold_cereals_by_calories",
        "model.dagster_dbt_test_project.sort_hot_cereals_by_calories",
    }

    def test_run_all(self, dbt_rpc_server):  # pylint: disable=unused-argument
        run_all_fast_poll = configured(dbt_rpc_run_and_wait, name="run_all_fast_poll")(
            {"interval": 2}
        )

        dagster_result, dbt_result = output_for_solid_executed_with_rpc_resource(run_all_fast_poll)

        executed_model_from_result = set(res.node["unique_id"] for res in dbt_result.results)
        assert executed_model_from_result == TestDBTRunAndWaitSolid.ALL_MODELS_KEY_SET

        materialization_asset_keys = set(
            mat.asset_key.to_string() for mat in dagster_result.materializations_during_compute
        )
        assert materialization_asset_keys == TestDBTRunAndWaitSolid.ALL_MODELS_KEY_SET

    SINGLE_MODEL_KEY_SET = {"model.dagster_dbt_test_project.least_caloric"}

    def test_run_single(self, dbt_rpc_server):  # pylint: disable=unused-argument
        run_single_fast_poll = configured(dbt_rpc_run_and_wait, name="run_least_caloric")(
            {"interval": 2, "models": ["least_caloric"]}
        )

        dagster_result, dbt_result = output_for_solid_executed_with_rpc_resource(
            run_single_fast_poll
        )

        executed_model_from_result = set(res.node["unique_id"] for res in dbt_result.results)
        assert executed_model_from_result == TestDBTRunAndWaitSolid.SINGLE_MODEL_KEY_SET

        materialization_asset_keys = set(
            mat.asset_key.to_string() for mat in dagster_result.materializations_during_compute
        )
        assert materialization_asset_keys == TestDBTRunAndWaitSolid.SINGLE_MODEL_KEY_SET


SINGLE_OP_CONFIGS: Dict[str, Tuple[SolidDefinition, Dict[str, Any]]] = {
    "run": (dbt_rpc_run, {"models": ["model_1", "model_2"], "exclude": ["bad_model"]}),
    "test": (
        dbt_rpc_test,
        {
            "models": ["model_1", "model_2"],
            "exclude": ["bad_model"],
            "data": False,
            "schema": True,
        },
    ),
}


class TestDBTSingleOperationSolids:
    @pytest.mark.parametrize("op", list(SINGLE_OP_CONFIGS.keys()))
    def test_dbt_rpc_single_op(self, op: str):
        op_solid, op_config = SINGLE_OP_CONFIGS[op]

        mocked_rpc_client = MagicMock(spec=DbtRpcClient)
        mocked_client_op_method = getattr(mocked_rpc_client, op)

        @resource
        def mock_dbt_rpc_resource(_init_context):
            return mocked_rpc_client

        response_sentinel_value = "<rpc response: {}>".format(uuid.uuid4())
        mocked_client_op_method.return_value.text = response_sentinel_value
        request_token_sentinel_value = "<request token: {}>".format(uuid.uuid4())
        mocked_client_op_method.return_value.json.return_value = {
            "result": {"request_token": request_token_sentinel_value}
        }

        configured_solid = configured(op_solid, name="configured_solid")(op_config)

        instance = DagsterInstance.ephemeral()
        result = execute_pipeline(
            PipelineDefinition(
                [configured_solid],
                mode_defs=[ModeDefinition(resource_defs={"dbt_rpc": mock_dbt_rpc_resource})],
            ),
            instance=instance,
        )

        mocked_client_op_method.assert_called_once_with(**op_config)
        assert (
            result.output_for_solid(configured_solid.name, "request_token")
            == request_token_sentinel_value
        )
        assert any(
            response_sentinel_value in event.message for event in instance.all_logs(result.run_id)
        )



import json
import platform
import sys
import uuid
from base64 import standard_b64encode as b64
from typing import Any, Dict, List, Optional

import attr
import requests

from dagster import Field, IntSource, RetryRequested, StringSource, resource

from .utils import is_fatal_code


@attr.s
class DbtRpcClient(object):

    host: str = attr.ib(default="0.0.0.0")
    port: int = attr.ib(validator=attr.validators.instance_of(int), default=8580)
    jsonrpc_version: str = attr.ib(validator=attr.validators.instance_of(str), default="2.0")
    url: str = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.url = f"http://{self.host}:{self.port}/jsonrpc"

    @staticmethod
    def _construct_user_agent() -> str:
        """Constructs a standard User Agent string to be used in headers for HTTP requests."""
        client = "dagster/dbt-rpc-client"
        python_version = (
            f"Python/{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        )
        system_info = f"{platform.system()}/{platform.release()}"
        user_agent = " ".join([python_version, client, system_info])
        return user_agent

    def _construct_headers(self) -> Dict:
        """Constructs a standard set of headers for HTTP requests."""
        headers = requests.utils.default_headers()
        headers["User-Agent"] = self._construct_user_agent()
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
        return headers

    def _post(self, data: str = None) -> requests.Response:
        """Constructs a standard way of making a POST request to the dbt RPC server."""
        headers = self._construct_headers()
        try:
            response = requests.post(self.url, headers=headers, data=data)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if is_fatal_code(e):
                raise e
            else:
                raise RetryRequested(max_retries=5, seconds_to_wait=30)  # TODO backoff logic
        return response

    def _default_request(self, method: str) -> Dict:
        data = {
            "jsonrpc": self.jsonrpc_version,
            "method": method,
            "id": str(uuid.uuid1()),
            "params": {},
        }
        return data

    def _selection(
        self, *, models: List[str] = None, select: List[str] = None, exclude: List[str] = None
    ) -> Dict:
        params = {}
        if models is not None:
            params["models"] = " ".join(set(models))
        if select is not None:
            params["select"] = " ".join(set(select))
        if exclude is not None:
            params["exclude"] = " ".join(set(exclude))

        return params

    def status(self) -> requests.Response:
        data = self._default_request(method="status")
        return self._post(data=json.dumps(data))

    def poll(
        self, *, request_token: str, logs: bool = False, logs_start: int = 0
    ) -> requests.Response:
        data = self._default_request(method="poll")
        data["params"] = {"request_token": request_token, "logs": logs, "logs_start": logs_start}
        return self._post(data=json.dumps(data))

    def ps(self, *, completed: bool = False) -> requests.Response:
        "Lists running and completed processes executed by the RPC server."
        data = self._default_request(method="ps")
        data["params"] = {"completed": completed}
        return self._post(data=json.dumps(data))

    def kill(self, *, task_id: str) -> requests.Response:
        "Terminates a running RPC task."
        data = self._default_request(method="kill")
        data["params"] = {"task_id": task_id}
        return self._post(data=json.dumps(data))

    def cli(self, *, cli: str, **kwargs) -> requests.Response:
        "Terminates a running RPC task."
        data = self._default_request(method="cli_args")
        data["params"] = {"cli": cli}

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def compile(
        self, *, models: List[str] = None, exclude: List[str] = None, **kwargs
    ) -> requests.Response:
        "Runs a dbt compile command."
        data = self._default_request(method="compile")
        data["params"].update(self._selection(models=models, exclude=exclude))

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def run(
        self, *, models: List[str] = None, exclude: List[str] = None, **kwargs
    ) -> requests.Response:
        "Runs a dbt run command."
        data = self._default_request(method="run")
        data["params"].update(self._selection(models=models, exclude=exclude))

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def snapshot(
        self, *, select: List[str] = None, exclude: List[str] = None, **kwargs
    ) -> requests.Response:
        "Runs a dbt snapshot command."
        data = self._default_request(method="snapshot")
        data["params"].update(self._selection(select=select, exclude=exclude))

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def test(
        self,
        *,
        models: List[str] = None,
        exclude: List[str] = None,
        data: bool = True,
        schema: bool = True,
        **kwargs
    ) -> requests.Response:
        payload = self._default_request(method="test")
        payload["params"] = {"data": data, "schema": schema}

        payload["params"].update(self._selection(models=models, exclude=exclude))

        if kwargs is not None:
            payload["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(payload))

    def seed(self, *, show: bool = False, **kwargs) -> requests.Response:
        data = self._default_request(method="seed")
        data["params"] = {"show": show}

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def generate_docs(
        self,
        *,
        models: List[str] = None,
        exclude: List[str] = None,
        compile: bool = False,  # pylint: disable=redefined-builtin # TODO
        **kwargs
    ) -> requests.Response:
        data = self._default_request(method="docs.generate")
        data["params"] = {"compile": compile}

        data["params"].update(self._selection(models=models, exclude=exclude))

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def run_operation(
        self, *, macro: str, args: Optional[Dict[str, Any]] = None, **kwargs
    ) -> requests.Response:
        data = self._default_request(method="run-operation")
        data["params"] = {"macro": macro}

        if args is not None:
            data["params"]["args"] = args

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def snapshot_freshness(
        self, *, select: Optional[List[str]] = None, **kwargs
    ) -> requests.Response:
        data = self._default_request(method="snapshot-freshness")
        data["params"].update(self._selection(select=select))

        if kwargs is not None:
            data["params"]["task_tags"] = kwargs

        return self._post(data=json.dumps(data))

    def compile_sql(self, *, sql: str, name: str) -> requests.Response:
        data = self._default_request(method="compile_sql")
        data["params"] = {"sql": b64(sql.encode("utf-8")).decode("utf-8"), "name": name}
        return self._post(data=json.dumps(data))

    def run_sql(self, *, sql: str, name: str) -> requests.Response:
        data = self._default_request(method="run_sql")
        data["params"] = {"sql": b64(sql.encode("utf-8")).decode("utf-8"), "name": name}
        return self._post(data=json.dumps(data))


@resource(
    description="A resource representing a dbt RPC client.",
    config_schema={
        "host": Field(StringSource),
        "port": Field(IntSource, is_required=False, default_value=8580),
    },
)
def dbt_rpc_resource(context) -> DbtRpcClient:
    return DbtRpcClient(host=context.resource_config["host"], port=context.resource_config["port"])


local_dbt_rpc_resource = dbt_rpc_resource.configured({"host": "0.0.0.0", "port": 8580})



import logging
from collections import defaultdict
from enum import Enum
from typing import Dict, List

from requests import Response
from requests.exceptions import RequestException

from dagster import EventMetadataEntry, Failure, RetryRequested
from dagster.core.execution.context.compute import SolidExecutionContext


def fmt_rpc_logs(logs: List[Dict]) -> Dict[int, str]:
    d = defaultdict(list)
    for log in logs:
        levelname = log.get("levelname")
        d[getattr(logging, levelname)].append(
            f"{log.get('timestamp')} - {levelname} - {log.get('message')}"
        )

    return {level: "\n".join(logs) for level, logs in d.items()}


def log_rpc(context: SolidExecutionContext, logs: List[Dict]) -> None:
    if len(logs) > 0:
        logs_fmt = fmt_rpc_logs(logs)
        for level, logs_str in logs_fmt.items():
            context.log.log(level=level, msg=logs_str)


class DBTErrors(Enum):
    project_currently_compiling_error = 10010
    runtime_error = 10001
    server_error = -32000
    project_compile_failure_error = 10011
    rpc_process_killed_error = 10009
    rpc_timeout_error = 10008


def raise_for_rpc_error(context: SolidExecutionContext, resp: Response) -> None:
    error = resp.json().get("error")
    if error is not None:
        if error["code"] in [
            DBTErrors.project_currently_compiling_error,
            DBTErrors.runtime_error,
            DBTErrors.server_error,
        ]:
            context.log.warning(error["message"])
            raise RetryRequested(max_retries=5, seconds_to_wait=30)  # TODO backoff logic
        elif error["code"] == DBTErrors.project_compile_failure_error:
            raise Failure(
                description=error["message"],
                metadata_entries=[
                    EventMetadataEntry.text(text=str(error["code"]), label="RPC Error Code"),
                    EventMetadataEntry.text(
                        text=error["data"]["cause"]["message"], label="RPC Error Cause"
                    ),
                ],
            )
        elif error["code"] == DBTErrors.rpc_process_killed_error:
            raise Failure(
                description=error["message"],
                metadata_entries=[
                    EventMetadataEntry.text(text=str(error["code"]), label="RPC Error Code"),
                    EventMetadataEntry.text(text=str(error["data"]["signum"]), label="RPC Signum"),
                    EventMetadataEntry.text(
                        text=error["data"]["message"], label="RPC Error Message"
                    ),
                ],
            )
        elif error["code"] == DBTErrors.rpc_timeout_error:
            raise Failure(
                description=error["message"],
                metadata_entries=[
                    EventMetadataEntry.text(text=str(error["code"]), label="RPC Error Code"),
                    EventMetadataEntry.text(
                        text=str(error["data"]["timeout"]), label="RPC Timeout"
                    ),
                    EventMetadataEntry.text(
                        text=error["data"]["message"], label="RPC Error Message"
                    ),
                ],
            )
        else:
            raise Failure(
                description=error["message"],
                metadata_entries=[
                    EventMetadataEntry.text(text=str(error["code"]), label="RPC Error Code"),
                ],
            )


def is_fatal_code(e: RequestException) -> bool:
    """Helper function to determine if a Requests reponse status code
    is a "fatal" status code. If it is, we will not request a solid retry."""
    return 400 <= e.response.status_code < 500 and e.response.status_code != 429
